
#include "USBCommunication.h"
#include <setupapi.h>
#include <string.h>
#include <stdio.h>


//#define MY_DEVICE_ID  "vid_04d8&pid_0060"

//Need to get a write "handle" to our device before we can write to it.
static HANDLE WriteHandle = INVALID_HANDLE_VALUE;

//Need to get a read "handle" to our device before we can read from it.
static HANDLE ReadHandle = INVALID_HANDLE_VALUE;

//The last error message
static char lastError[255];

//The device ID
static char lastDeviceID[255];

//Signal that the communication component was initialised
static BOOL initialised = FALSE;

BOOL initUSBCommunication(const char* deviceID)
{
	BOOL ret = FALSE;

	GUID InterfaceClassGuid = { 0x4d1e55b2, 0xf16f, 0x11cf, { 0x88, 0xcb, 0x00, 0x11, 0x11, 0x00, 0x00, 0x30 } };

	HDEVINFO DeviceInfoTable = INVALID_HANDLE_VALUE;
	PSP_DEVICE_INTERFACE_DATA
			InterfaceDataStructure = (PSP_DEVICE_INTERFACE_DATA) malloc(
					sizeof(SP_DEVICE_INTERFACE_DATA));
	PSP_DEVICE_INTERFACE_DETAIL_DATA DetailedInterfaceDataStructure =
			(PSP_DEVICE_INTERFACE_DETAIL_DATA) malloc(
					sizeof(SP_DEVICE_INTERFACE_DETAIL_DATA));
	SP_DEVINFO_DATA DevInfoData;

	DWORD InterfaceIndex = 0;
	DWORD dwRegType;
	DWORD dwRegSize;
	DWORD StructureSize = 0;
	PBYTE PropertyValueBuffer = NULL;
	DWORD ErrorStatus;

	//First populate a list of plugged in devices (by specifying "DIGCF_PRESENT"), which are of the specified class GUID.
	DeviceInfoTable = SetupDiGetClassDevs(&InterfaceClassGuid, NULL, NULL,
			DIGCF_PRESENT | DIGCF_DEVICEINTERFACE);

	//Now look through the list we just populated.  We are trying to see if any of them match our device.
	while (TRUE) {
		InterfaceDataStructure->cbSize = sizeof(SP_DEVICE_INTERFACE_DATA);
		if (SetupDiEnumDeviceInterfaces(DeviceInfoTable, NULL,
				&InterfaceClassGuid, InterfaceIndex, InterfaceDataStructure)) {
			ErrorStatus = GetLastError();
			if (ERROR_NO_MORE_ITEMS == ErrorStatus) //Did we reach the end of the list of matching devices in the DeviceInfoTable?
			{ //Cound not find the device.  Must not have been attached.
				SetupDiDestroyDeviceInfoList(DeviceInfoTable); //Clean up the old structure we no longer need.
				strcpy(lastError,"USB device not connected or Device ID incorrect!");
				break;
			}
		} else //Else some other kind of unknown error ocurred...
		{
			ErrorStatus = GetLastError();
			sprintf(lastError, "Execution ended unexpected! Error code: %d", (int)ErrorStatus);
			SetupDiDestroyDeviceInfoList(DeviceInfoTable); //Clean up the old structure we no longer need.
			break;
		}

		//Now retrieve the hardware ID from the registry.  The hardware ID contains the VID and PID, which we will then
		//check to see if it is the correct device or not.

		//Initialize an appropriate SP_DEVINFO_DATA structure.  We need this structure for SetupDiGetDeviceRegistryProperty().
		DevInfoData.cbSize = sizeof(SP_DEVINFO_DATA);
		SetupDiEnumDeviceInfo(DeviceInfoTable, InterfaceIndex, &DevInfoData);

		//First query for the size of the hardware ID, so we can know how big a buffer to allocate for the data.
		SetupDiGetDeviceRegistryProperty(DeviceInfoTable, &DevInfoData,
				SPDRP_HARDWAREID, &dwRegType, NULL, 0, &dwRegSize);

		//Allocate a buffer for the hardware ID.
		if (PropertyValueBuffer != NULL) {
			free(PropertyValueBuffer);
		}
		PropertyValueBuffer = (BYTE *) malloc(dwRegSize);
		if (PropertyValueBuffer == NULL) //if null, error, couldn't allocate enough memory
		{ //Can't really recover from this situation, just exit instead.
			SetupDiDestroyDeviceInfoList(DeviceInfoTable); //Clean up the old structure we no longer need.
			strcpy(lastError, "Cannot allocate memory !");
			break;
		}

		//Retrieve the hardware IDs for the current device we are looking at.  PropertyValueBuffer gets filled with a
		//REG_MULTI_SZ (array of null terminated strings).  To find a device, we only care about the very first string in the
		//buffer, which will be the "device ID".  The device ID is a string which contains the VID and PID, in the example
		//format "Vid_04d8&Pid_003f".
		SetupDiGetDeviceRegistryProperty(DeviceInfoTable, &DevInfoData,
				SPDRP_HARDWAREID, &dwRegType, PropertyValueBuffer, dwRegSize,
				NULL);

		//Now check if the hardware ID we are looking at contains the correct VID/PID
		if (strstr(strlwr((char*)PropertyValueBuffer),deviceID) != NULL) {
			//Device must have been found.  Open read and write handles.  In order to do this, we will need the actual device path first.
			//We can get the path by calling SetupDiGetDeviceInterfaceDetail(), however, we have to call this function twice:  The first
			//time to get the size of the required structure/buffer to hold the detailed interface data, then a second time to actually
			//get the structure (after we have allocated enough memory for the structure.)
			DetailedInterfaceDataStructure->cbSize
					= sizeof(SP_DEVICE_INTERFACE_DETAIL_DATA);
			//First call populates "StructureSize" with the correct value
			SetupDiGetDeviceInterfaceDetail(DeviceInfoTable, InterfaceDataStructure, NULL, 0, &StructureSize, NULL);
			if (DetailedInterfaceDataStructure != NULL) {
				free(DetailedInterfaceDataStructure);
			}
			DetailedInterfaceDataStructure
					= (PSP_DEVICE_INTERFACE_DETAIL_DATA) (malloc(StructureSize)); //Allocate enough memory
			if (DetailedInterfaceDataStructure == NULL) //if null, error, couldn't allocate enough memory
			{ //Can't really recover from this situation, just exit instead.
				SetupDiDestroyDeviceInfoList(DeviceInfoTable); //Clean up the old structure we no longer need.
				strcpy(lastError, "Cannot allocate memory !");
				break;
			}
			DetailedInterfaceDataStructure->cbSize
					= sizeof(SP_DEVICE_INTERFACE_DETAIL_DATA);
			//Now call SetupDiGetDeviceInterfaceDetail() a second time to receive the goods.
			SetupDiGetDeviceInterfaceDetail(DeviceInfoTable,
					InterfaceDataStructure, DetailedInterfaceDataStructure,
					StructureSize, NULL, NULL);

			//We now have the proper device path, and we can finally open read and write handles to the device.
			//We store the handles in the global variables "WriteHandle" and "ReadHandle", which we will use later to actually communicate.
			WriteHandle = CreateFile(
					(DetailedInterfaceDataStructure->DevicePath),
					GENERIC_WRITE, FILE_SHARE_READ | FILE_SHARE_WRITE, NULL,
					OPEN_EXISTING, 0, 0);

			ReadHandle = CreateFile(
					(DetailedInterfaceDataStructure->DevicePath), GENERIC_READ,
					FILE_SHARE_READ | FILE_SHARE_WRITE, NULL, OPEN_EXISTING, 0,
					0);

			SetupDiDestroyDeviceInfoList(DeviceInfoTable); //Clean up the old structure we no longer need.
			ret = TRUE;
			//save the device ID and signal that the component is initialised
			strcpy(lastDeviceID, deviceID);
			lastDeviceID[strlen(deviceID)] = 0;
			initialised = TRUE;
			break;
		}

		InterfaceIndex++;
		//Keep looping until we either find a device with matching VID and PID, or until we run out of items.
	}//end of while(true)

	if (PropertyValueBuffer != NULL) {
		free(PropertyValueBuffer);
	}
	free(InterfaceDataStructure);
	free(DetailedInterfaceDataStructure);

	return ret;
}

char sendCommand(char command)
{
    DWORD BytesWritten = 0;
    DWORD BytesRead = 0;
    unsigned char OutputPacketBuffer[65];  //Allocate a memory buffer equal to our endpoint size + 1
    unsigned char InputPacketBuffer[65];  //Allocate a memory buffer equal to our endpoint size + 1
    BOOL opResult;

    //Initialize the connection
    if (!initialised) {
    	opResult = initUSBCommunication(lastDeviceID);
    	if(opResult == FALSE) {
    		initialised = FALSE;
    		/*return the closed state*/
    		return 0x02;

    	}
    }

    InputPacketBuffer[0] = 0;        //The first byte is the "Report ID" and does not get transmitted over the USB bus.  Always set = 0.
    OutputPacketBuffer[0] = 0;        //The first byte is the "Report ID" and does not get transmitted over the USB bus.  Always set = 0.

    OutputPacketBuffer[1] = 0x80;      //0x80 is the "send command" command in the firmware
    OutputPacketBuffer[2] = (unsigned char)command;      //the command

    //The basic Windows I/O functions WriteFile() and ReadFile() can be used to read and write to HID class USB devices
    //(once we have the read and write handles to the device, which are obtained with CreateFile()).

    //The following call to WriteFile() sends 64 bytes of data to the USB device.
    opResult = WriteFile(WriteHandle, &OutputPacketBuffer, 65, &BytesWritten, 0);  //Blocking function, unless an "overlapped" structure is used
    if (!opResult) {
    	/*close the handles*/
    	CloseHandle(WriteHandle);
    	CloseHandle(ReadHandle);

		initialised = FALSE;
		/*return the closed state*/
		return 0x02;
    }
    //Now get the response packet from the firmware.
    //The following call to ReadFIle() retrieves 64 bytes of data from the USB device.
    opResult = ReadFile(ReadHandle, &InputPacketBuffer, 65, &BytesRead, 0);    //Blocking function, unless an "overlapped" structure is used
    if (!opResult) {
    	/*close the handles*/
    	CloseHandle(WriteHandle);
    	CloseHandle(ReadHandle);

    	initialised = FALSE;
		/*return the closed state*/
		return 0x02;
    }

    //InputPacketBuffer[0] is the report ID, which we don't care about.
    //InputPacketBuffer[1] contains the state.


	/*return the state*/
	return (char)InputPacketBuffer[1];
}


char* sendText(const char* text, int size)
{
    DWORD BytesWritten = 0;
    DWORD BytesRead = 0;
    unsigned char OutputPacketBuffer[65];  //Allocate a memory buffer equal to our endpoint size + 1
    unsigned char InputPacketBuffer[65];  //Allocate a memory buffer equal to our endpoint size + 1
    BOOL opResult;

    //Initialize the connection
    if (!initialised) {
    	opResult = initUSBCommunication(lastDeviceID);
    	if(opResult == FALSE) {
    		initialised = FALSE;
    		/*return the closed state*/
    		return 0x02;

    	}
    }

    InputPacketBuffer[0] = 0;        //The first byte is the "Report ID" and does not get transmitted over the USB bus.  Always set = 0.
    OutputPacketBuffer[0] = 0;        //The first byte is the "Report ID" and does not get transmitted over the USB bus.  Always set = 0.

    memcpy(&OutputPacketBuffer[1], text, size);

    //The basic Windows I/O functions WriteFile() and ReadFile() can be used to read and write to HID class USB devices
    //(once we have the read and write handles to the device, which are obtained with CreateFile()).

    //The following call to WriteFile() sends 64 bytes of data to the USB device.
    opResult = WriteFile(WriteHandle, &OutputPacketBuffer, 65, &BytesWritten, 0);  //Blocking function, unless an "overlapped" structure is used
    if (!opResult) {
    	/*close the handles*/
    	CloseHandle(WriteHandle);
    	CloseHandle(ReadHandle);

		initialised = FALSE;
		/*return the closed state*/
		return 0x02;
    }
    //Now get the response packet from the firmware.
    //The following call to ReadFIle() retrieves 64 bytes of data from the USB device.
    opResult = ReadFile(ReadHandle, &InputPacketBuffer, 65, &BytesRead, 0);    //Blocking function, unless an "overlapped" structure is used
    if (!opResult) {
    	/*close the handles*/
    	CloseHandle(WriteHandle);
    	CloseHandle(ReadHandle);

    	initialised = FALSE;
		/*return the closed state*/
		return 0x02;
    }

    //InputPacketBuffer[0] is the report ID, which we don't care about.
    //InputPacketBuffer[1] contains the state.

	printf("%s\n", (char*)&InputPacketBuffer[1]);
	
	/*return the state*/
	return (char*)&InputPacketBuffer[1];
}

char getPositions(double* positionLeft, double* positionRight)
{
    DWORD BytesWritten = 0;
    DWORD BytesRead = 0;
    unsigned char OutputPacketBuffer[65];  //Allocate a memory buffer equal to our endpoint size + 1
    unsigned char InputPacketBuffer[65];  //Allocate a memory buffer equal to our endpoint size + 1
    float posLeft;
    float posRight;
    BOOL opResult;

    //Initialize the connection
    if (!initialised) {
    	opResult = initUSBCommunication(lastDeviceID);
    	if(opResult == FALSE) {
    		initialised = FALSE;
    		/*return the closed state*/
    		return 0x02;

    	}
    }

    InputPacketBuffer[0] = 0;        //The first byte is the "Report ID" and does not get transmitted over the USB bus.  Always set = 0.
    OutputPacketBuffer[0] = 0;        //The first byte is the "Report ID" and does not get transmitted over the USB bus.  Always set = 0.

    OutputPacketBuffer[1] = 0x81;      //0x81 is the "get positions" command in the firmware

    //The basic Windows I/O functions WriteFile() and ReadFile() can be used to read and write to HID class USB devices
    //(once we have the read and write handles to the device, which are obtained with CreateFile()).

    //To get the pushbutton state, first, we send a packet with our "Get Pushbutton State" command in it.
    //The following call to WriteFile() sends 64 bytes of data to the USB device.
    opResult = WriteFile(WriteHandle, &OutputPacketBuffer, 65, &BytesWritten, 0);  //Blocking function, unless an "overlapped" structure is used
    if (!opResult) {
    	/*close the handles*/
    	CloseHandle(WriteHandle);
    	CloseHandle(ReadHandle);

		initialised = FALSE;
		/*return the closed state*/
		*positionLeft = 1.0;
		*positionRight = 1.0;
		return 0x02;
    }

    //Now get the response packet from the firmware.
    //The following call to ReadFIle() retrieves 64 bytes of data from the USB device.
    opResult = ReadFile(ReadHandle, &InputPacketBuffer, 65, &BytesRead, 0);    //Blocking function, unless an "overlapped" structure is used
    if (!opResult) {
    	/*close the handles*/
    	CloseHandle(WriteHandle);
    	CloseHandle(ReadHandle);

		initialised = FALSE;
		/*return the closed state*/
		*positionLeft = 1.0;
		*positionRight = 1.0;
		return 0x02;
    }

    //InputPacketBuffer[0] is the report ID, which we don't care about.
    //InputPacketBuffer[1] contains the state.

    //InputPacketBuffer[2]
    //InputPacketBuffer[3]
    //InputPacketBuffer[4]
    //InputPacketBuffer[5] contains the left position as a float
	memcpy(&posLeft, &InputPacketBuffer[2], 4);

    //InputPacketBuffer[6]
    //InputPacketBuffer[7]
    //InputPacketBuffer[8]
    //InputPacketBuffer[9] contains the left position as a float
	memcpy(&posRight, &InputPacketBuffer[6], 4);

	*positionLeft = posLeft;
	*positionRight = posRight;

	return (char)InputPacketBuffer[1];
}

char* getLastError() {
	return lastError;
}
